package be.iminds.ilabt.jfed.experimenter_gui.renew_report;

import be.iminds.ilabt.jfed.JFedWebApiClientModule;
import be.iminds.ilabt.jfed.bugreport.BugReport;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportComposer;
import be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportToServerTask;
import be.iminds.ilabt.jfed.experimenter_gui.ui.SceneAwareController;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClientConfig;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.TimeUtil;
import com.google.common.base.Throwables;
import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.inject.name.Named;
import javafx.application.Platform;
import javafx.beans.binding.Binding;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.property.ReadOnlyDoubleProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyCodeCombination;
import javafx.scene.layout.VBox;
import javafx.scene.text.Font;
import javafx.scene.text.FontWeight;
import javafx.scene.layout.RowConstraints;
import javafx.stage.Stage;
import org.controlsfx.dialog.ExceptionDialog;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Provider;
import java.net.URL;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

public class RenewReportDialog implements SceneAwareController, Initializable {
    private Slice slice;
    private ExperimentTaskStatus jobSuccessExperimentTaskStatus;
    private ExperimentController experimentController;
    private Throwable jobFailureThrowable;
    private Instant requestedRenewInstant;

    @Nonnull private final RenewReportDialogFactory renewReportDialogFactory;
    @Nonnull private final Provider<BugReportComposer> bugReportComposerProvider;
    @Nonnull private final Provider<TaskService> taskServiceProvider;
    @Nonnull private final Provider<FedmonWebApiClientConfig> fedmonWebApiClientConfigProvider;

    @FXML private ProgressBar busyBar;

    @FXML private RowConstraints failInfoRow;

    @FXML private Label helpLabel;

    @FXML private Label summaryStatusLabel;
    @FXML private Label summaryTargetExpirationLabel;
    @FXML private Label summaryFirstExpirationLabel;
    @FXML private Label sliceExpirationLabel;
    @FXML private Label earliestSliverExpirationLabel;

    @FXML private TableColumn amColumn;
    @FXML private TableColumn sliverColumn;
    @FXML private TableColumn expirationColumn;

    @FXML private Button okButton;
    @FXML private Button bugreportButton;
    @FXML private Button retryButton;
    @FXML private Button refreshButton;

    @FXML private TableView<ExpirationDetail> expirationTable;
    private final ObservableList<ExpirationDetail> expirationDetails;

    private Scene scene;
    private Stage stage;

    @Inject
    public RenewReportDialog(@Nonnull RenewReportDialogFactory renewReportDialogFactory,
                             @Nonnull Provider<BugReportComposer> bugReportComposerProvider,
                             @Nonnull Provider<TaskService> taskServiceProvider,
                             @Nonnull @Named(JFedWebApiClientModule.NAMED__CLIENT_AUTH_WEBAPI_CONFIG) Provider<FedmonWebApiClientConfig> fedmonWebApiClientConfigProvider) {
        this.renewReportDialogFactory = renewReportDialogFactory;
        this.bugReportComposerProvider = bugReportComposerProvider;
        this.taskServiceProvider = taskServiceProvider;
        this.fedmonWebApiClientConfigProvider = fedmonWebApiClientConfigProvider;
        this.expirationDetails = FXCollections.observableArrayList();
    }

    @Override
    public void initialize(URL url, ResourceBundle resourceBundle) {
        expirationTable.setItems(expirationDetails);

        //populate table
        ReadOnlyDoubleProperty dividableWidth = expirationTable.widthProperty();
        amColumn.prefWidthProperty().bind(dividableWidth.multiply(0.25));
        sliverColumn.prefWidthProperty().bind(dividableWidth.multiply(0.375));
        expirationColumn.prefWidthProperty().bind(dividableWidth.multiply(0.375));
    }

    @Override
    public void registerScene(Scene scene) {
        this.scene = scene;
        this.stage = (Stage) scene.getWindow();

        scene.getAccelerators().put(new KeyCodeCombination(KeyCode.F10), () -> simulateFailForDebug());
    }


    public class ExpirationDetail {
        private final String am;
        private final String sliver;
        private final String expiration;
        private final boolean ok;

        public ExpirationDetail(String am, String sliver, Instant expiration, boolean ok) {
            this(am, sliver, toHumanTime(expiration), ok);
        }
        public ExpirationDetail(String am, String sliver, String expiration, boolean ok) {
            this.am = am;
            this.sliver = sliver;
            this.expiration = expiration;
            this.ok = ok;
        }

        public String getAm() {
            return am;
        }

        public String getSliver() {
            return sliver;
        }

        public String getExpiration() {
            return expiration;
        }

        public boolean isOk() {
            return ok;
        }
    }

    public void closeDialog() {
        if (this.stage != null)
            this.stage.close();
    }
    public void retryRenew() {
        ListenableFuture<ExperimentTaskStatus> future = experimentController.renew(requestedRenewInstant);

        Futures.addCallback(future, new FutureCallback<ExperimentTaskStatus>() {
            @Override
            public void onSuccess(@Nullable ExperimentTaskStatus experimentTaskStatus) {
                Platform.runLater(() -> {
                    renewReportDialogFactory.showDialog(requestedRenewInstant, experimentController, experimentTaskStatus);
                });
            }
            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                Platform.runLater(() -> {
                    renewReportDialogFactory.showDialog(requestedRenewInstant, experimentController, throwable);
                });
            }
        });
        if (this.stage != null)
            this.stage.close();
    }
    public void sendBugreport() {
        Alert userInfoDialog = new Alert(Alert.AlertType.CONFIRMATION);
        if (stage != null)
            userInfoDialog.initOwner(stage);
        userInfoDialog.setHeaderText("Extra Info");
        userInfoDialog.setResizable(true);

        VBox contentPane = new VBox(10);
        contentPane.setPrefWidth(600);
        Label headerLabel = new Label("How important/urgent is this Renew?");
        headerLabel.setFont(Font.font(Font.getDefault().getFamily(), FontWeight.BOLD, Font.getDefault().getSize()));

        Label infoLabel = new Label("Could you please provide info to help us set a priority " +
                "for this bug report?");
        infoLabel.setWrapText(true);

        TextArea descriptionTextArea = new TextArea();
        descriptionTextArea.setPromptText("add extra info here");
        descriptionTextArea.setStyle("-fx-margin: 0 0 0 30px");
        descriptionTextArea.setPrefRowCount(3);

        contentPane.getChildren().addAll(headerLabel, infoLabel, descriptionTextArea);
        userInfoDialog.getDialogPane().setContent(contentPane);

        Platform.runLater(() -> {
            if (userInfoDialog.showAndWait().isPresent()) {
                String userInfo = descriptionTextArea.getText();

                BugReportComposer bugReportComposer = bugReportComposerProvider.get();
                bugReportComposer.setFirstRunConnectivityTests(false);

                String subject = "Renew Failed for "+slice.getUrn().getEncodedTopLevelAuthority_withoutSubAuth()+" slice \""+slice.getUrn().getEncodedResourceName()+"\"";

                bugReportComposer.setPostOnPublicList(false);
                bugReportComposer.setReportType(BugReport.ReportType.BUG);
                bugReportComposer.setReportTarget(BugReport.ReportTarget.OTHER);
                bugReportComposer.setSubject(subject);

                String description = "";

                description += "Slice URN: "+slice.getUrnString()+"\n";
                description += "Renew Summary: "+summaryStatusLabel.getText()+"\n";
                description += "Requested expiration time: "+summaryTargetExpirationLabel.getText()+"\n";
                description += "Actual expiration time: "+summaryFirstExpirationLabel.getText()+"\n";
                description += "\n";

                description += "Slice expiration time: "+sliceExpirationLabel.getText()+"\n";
                description += "Earliest Sliver expiration time: "+earliestSliverExpirationLabel.getText()+"\n";
                description += "\n";
                description += "Sliver expiration overview:\n";
                for (ExpirationDetail expirationDetail: expirationDetails) {
                    description += "   - "+expirationDetail.getSliver()+" expires at "+expirationDetail.getExpiration()+"\n";
                }

                if (userInfo != null && !userInfo.trim().isEmpty()) {
                    description += "\n\nExtra info from user:\n"+userInfo+"\n";
                }

                bugReportComposer.setDescription(description.toString());

                taskServiceProvider.get().submitTask(
                        new BugReportToServerTask(bugReportComposer, fedmonWebApiClientConfigProvider.get())
                );
            }
        });

        if (this.stage != null)
            this.stage.close();
    }
    public void refresh() {
        setBusy();
        ListenableFuture<Void> future = experimentController.requestUpdate();

        Futures.addCallback(future, new FutureCallback<Void>() {
            @Override
            public void onSuccess(@Nullable Void v) {
                Platform.runLater(() -> {
                    calculateStatus();
                });
            }
            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                Platform.runLater(() -> {
                    calculateStatus();
                });
            }
        });
    }

    public void setFailStatus(String text) {
        summaryStatusLabel.setText(text);
        setLabelStyle(summaryStatusLabel, false);
        bugreportButton.setVisible(true);
        retryButton.setVisible(true);
        okButton.setVisible(true);
        refreshButton.setVisible(true);
        helpLabel.setVisible(true);
        busyBar.setVisible(false);
        failInfoRow.setMinHeight(60);
        failInfoRow.setPrefHeight(100);
    }

    public void setSuccessStatus() {
        summaryStatusLabel.setText("Successfully Renewed Experiment");
        setLabelStyle(summaryStatusLabel, true);
        bugreportButton.setVisible(false);
        retryButton.setVisible(false);
        helpLabel.setVisible(false);
        okButton.setVisible(true);
        refreshButton.setVisible(true);
        busyBar.setVisible(false);
        failInfoRow.setMinHeight(0);
        failInfoRow.setPrefHeight(0);
    }

    public void setBusy() {
        summaryStatusLabel.setText("?");
        summaryTargetExpirationLabel.setText("?");
        summaryFirstExpirationLabel.setText("?");
        sliceExpirationLabel.setText("?");
        earliestSliverExpirationLabel.setText("?");

        setNoLabelStyle(summaryStatusLabel);
        setNoLabelStyle(summaryTargetExpirationLabel);
        setNoLabelStyle(summaryFirstExpirationLabel);
        setNoLabelStyle(sliceExpirationLabel);
        setNoLabelStyle(earliestSliverExpirationLabel);

        expirationDetails.clear();

        helpLabel.setVisible(false);
        bugreportButton.setVisible(false);
        retryButton.setVisible(false);
        okButton.setVisible(false);
        refreshButton.setVisible(false);
        busyBar.setVisible(true);
    }

    public void simulateFailForDebug() {
        //simulate a failed renew
//        requestedRenewInstant = requestedRenewInstant.plusSeconds(5*60*60); //add 5 hours
        requestedRenewInstant = requestedRenewInstant.plusSeconds(10*60); //add 10 minutes
        calculateStatus();
    }

    private static final DateTimeFormatter dataTimeFormatter =
            DateTimeFormatter.ofPattern( "yyyy/MM/dd  HH:mm:ss zzz" )
                    .withZone( ZoneId.systemDefault() );
    private static String toHumanTime(@Nullable Instant instant) {
        if (instant == null)
            return "unknown";

        return dataTimeFormatter.format(instant)+"  ("+ TimeUtil.calcIn(instant)+")";
    }

    @Contract("null -> true")
    private boolean isBeforeRequestedExpireTime(@Nullable Instant actualExpirationTime) {
        if (actualExpirationTime == null)
            return true;
        if (requestedRenewInstant == null)
            throw new IllegalStateException("requestedRenewInstant is null");

        //Request times are rounded down to the nearest second. We have to take that into account.
        //We also don't care about say 5 seconds, so we'll just ignore up to a 5 second difference.
        long actualMillis = actualExpirationTime.toEpochMilli();
        long targetMillis = requestedRenewInstant.toEpochMilli();
        return (actualMillis + 5000l) < targetMillis;
    }


    private void setLabelStyle(@Nonnull Label target, boolean ok) {
        if (ok) {
            target.getStyleClass().removeAll("label-text-error", "label-error");
            target.getStyleClass().addAll("label-ok", "label-text-ok");
        } else {
            target.getStyleClass().removeAll("label-ok", "label-text-ok");
            target.getStyleClass().addAll("label-text-error", "label-error");
        }
    }
    private void setNoLabelStyle(@Nonnull Label target) {
        target.getStyleClass().removeAll("label-text-error", "label-error", "label-ok", "label-text-ok");
    }

    /** Also sets style classes */
    private void setExpirationLabel(@Nonnull Label target,
                                    @Nullable Instant expireTime) {
        target.setText(toHumanTime(expireTime));
        setLabelStyle(target, !isBeforeRequestedExpireTime(expireTime));
    }

    /** this gets called after the setters, to update the status of the renew attempt */
    public void calculateStatus() {
        setBusy();

        if (experimentController == null)
            throw new IllegalStateException("experimentController was not set");
        if (requestedRenewInstant == null)
            throw new IllegalStateException("requestedRenewInstant was not set");
        if (slice == null)
            throw new IllegalStateException("slice was not set");

        summaryTargetExpirationLabel.setText(toHumanTime(requestedRenewInstant));
        setExpirationLabel(sliceExpirationLabel, slice.getExpirationDate());

        boolean allOk = !isBeforeRequestedExpireTime(slice.getExpirationDate());

        Instant earliestSliverExpiration = null;
        List<ExpirationDetail> newExpirationDetails = new ArrayList<>();
        for (Sliver sliver: slice.getSliversCopy()) {
            newExpirationDetails.add(new ExpirationDetail(
                    sliver.getServer().getName(),
                    sliver.getUrnString(),
                    sliver.getExpirationDate(),
                    !isBeforeRequestedExpireTime(sliver.getExpirationDate())
            ));
            if (sliver.getExpirationDate() != null) {
                if (earliestSliverExpiration == null ||
                        earliestSliverExpiration.isAfter(sliver.getExpirationDate()))
                    earliestSliverExpiration = sliver.getExpirationDate();
                allOk = allOk && !isBeforeRequestedExpireTime(slice.getExpirationDate());
            } else {
                allOk = false;
            }
        }
        setExpirationLabel(earliestSliverExpirationLabel, earliestSliverExpiration);
        Instant earliestSliverOrSliceExpiration =
                earliestSliverExpiration == null ?
                        slice.getExpirationDate() :
                        (slice.getExpirationDate() == null ? earliestSliverExpiration :
                                (slice.getExpirationDate().isBefore(earliestSliverExpiration) ?
                                        slice.getExpirationDate() : earliestSliverExpiration)
                        );
        setExpirationLabel(summaryFirstExpirationLabel, earliestSliverOrSliceExpiration);
        allOk = allOk && !isBeforeRequestedExpireTime(earliestSliverExpiration);

        expirationDetails.setAll(newExpirationDetails);

        //noinspection VariableNotUsedInsideIf
        if (jobFailureThrowable != null) {
            allOk = false;
            setFailStatus("Job Failed - Status Unknown");
            return;
        }
        if (jobSuccessExperimentTaskStatus == null) {
            allOk = false;
            setFailStatus("Job Status Unknown");
            return;
        }
        if (jobSuccessExperimentTaskStatus != ExperimentTaskStatus.SUCCESS) {
            allOk = false;
            setFailStatus("Job was not successfull: "+jobSuccessExperimentTaskStatus);
            return;
        }

        if (allOk) {
            setSuccessStatus();
        } else {
            setFailStatus("Renew failed.");
        }
    }

    public void setJobSuccessExperimentTaskStatus(@Nullable ExperimentTaskStatus jobSuccessExperimentTaskStatus) {
        this.jobSuccessExperimentTaskStatus = jobSuccessExperimentTaskStatus;
    }

    public void setExperimentController(@Nonnull ExperimentController experimentController) {
        this.experimentController = experimentController;
        this.slice = experimentController.getExperiment().getSlice();
    }

    public void setJobFailureThrowable(@Nonnull Throwable jobFailureThrowable) {
        this.jobFailureThrowable = jobFailureThrowable;
    }

    public void setRequestedRenewInstant(@Nonnull Instant requestedRenewInstant) {
        this.requestedRenewInstant = requestedRenewInstant;
    }
}
